from typing import Type
import warnings
import copy
from numbers import Integral, Real
import numpy as np
from sklearn.tree._criterion import Criterion, ClassificationCriterion
from sklearn.tree._splitter import Splitter
from sklearn.utils.validation import check_random_state
from sklearn.utils._param_validation import (
    validate_params, Interval,
)
from ._bipartite_splitter import BipartiteSplitter
from ._bipartite_criterion import GMOSA
from ._semisupervised_criterion import (
    SSCompositeCriterion,
    BipartiteSemisupervisedCriterion,
)
from ._axis_criterion import AxisCriterion
from . import _axis_criterion
from . import _unsupervised_criterion


AXIS_CRITERIA_CLF = {
    "gini": _axis_criterion.AxisGini,
    "entropy": _axis_criterion.AxisEntropy,
    "log_loss": _axis_criterion.AxisEntropy,
}
U_CATEGORIC_CRITERIA = {
    "gini": _unsupervised_criterion.UnsupervisedGini,
    "entropy": _unsupervised_criterion.UnsupervisedEntropy,
    "pairwise_gini": _unsupervised_criterion.PairwiseGini,
    "pairwise_entropy": _unsupervised_criterion.PairwiseEntropy,
}


def _duplicate_single_parameters(*params, ndim=2):
    result = []

    for param in params:
        if isinstance(param, (list, tuple)):
            if len(param) != ndim:
                raise ValueError(
                    f"param {param=} has length {len(param)}, expected {ndim}."
                )
            result.append(list(param))
        else:
            result.append([copy.deepcopy(param) for _ in range(ndim)])

    return result


def _is_classification_criterion_type(criterion: Type[Criterion]):
    return (
        criterion in AXIS_CRITERIA_CLF.values()
        or criterion in U_CATEGORIC_CRITERIA.values()
        or issubclass(criterion, ClassificationCriterion)
    )


def check_criterion(
    criterion: str | Criterion | Type[Criterion],
    *,
    n_samples: int | None = None,
    n_outputs: int | None = None,
    n_classes: int | None = None,
    kind: str | None = None,
    is_classification: bool | None = None,
    axis: bool = False,  # Useful only if isinstance(criterion, str)
) -> Criterion:

    if not isinstance(criterion, str) and kind is not None:
        warnings.warn(
            f"{kind=} parameter is being ignored since {criterion=} is "
            "not a string."
        )
    if isinstance(criterion, type):
        if not issubclass(criterion, (Criterion, AxisCriterion)):
            raise ValueError(
                f"Type {criterion=} provided is not a subclass of Criterion"
                " or AxisCriterion."
            )

        if is_classification is None:
            is_classification = _is_classification_criterion_type(criterion)

        if is_classification:
            if n_outputs is None or n_classes is None:
                raise ValueError(
                    "'n_outputs' and 'n_classes' must be provided if 'criterion' "
                    f"is a classification criterion type. Received {n_outputs=}, "
                    f"{n_classes=} and {criterion=}."
                )
            if len(n_classes) != n_outputs :
                raise ValueError(
                    "'len(n_classes)' must be 'n_outputs'. "
                    f"Received {n_outputs=} and {len(n_classes)=}. "
                    f"{n_classes=}."
                )
            final_criterion = criterion(
                n_outputs=n_outputs,
                n_classes=n_classes,
            )
        else:
            if n_outputs is None or n_samples is None:
                raise ValueError(
                    "'n_outputs' and 'n_samples' must be provided if 'criterion' "
                    f"is a regression criterion type. Received {n_outputs=}, "
                    f"{n_samples=} and {criterion=}."
                )
            final_criterion = criterion(
                n_outputs=n_outputs,
                n_samples=n_samples,
            )

        return final_criterion

    elif isinstance(criterion, str):
        raise NotImplementedError
    
    elif isinstance(criterion, Criterion):
        if n_samples is not None:
            warnings.warn(
                f"{n_samples=} parameter is being ignored since {criterion=} "
                "is already a Criterion instance."
            )
        if n_outputs is not None:
            warnings.warn(
                f"{n_outputs=} parameter is being ignored since {criterion=} "
                "is already a Criterion instance."
            )
        return criterion
    else:  # Should never run because of @validate_params
        raise ValueError(f"Unknown criterion type ({criterion=}).")


# TODO:
# @validate_params(
#     dict(
#         splitters=[list, tuple, Splitter],
#         criteria=[list, tuple, Criterion],
#         n_samples=[None, list, tuple, ],
#         max_features=None,
#         n_outputs=1,
#         min_samples_leaf=1,
#         min_weight_leaf=0.0,
#         ax_min_samples_leaf=1,
#         ax_min_weight_leaf=0.0,
#         random_state=None,
#         bipartite_criterion_class=GMOSA,
#     ),
#     prefer_skip_nested_validation=True,
# )
def make_bipartite_splitter(
    splitters,
    criteria,
    *,
    is_classification=None,
    n_samples=None,
    n_classes=None,
    n_outputs=1,
    max_features=None,
    min_samples_leaf=1,
    min_weight_leaf=0.0,
    ax_min_samples_leaf=1,
    ax_min_weight_leaf=0.0,
    random_state=None,
    bipartite_criterion_class=GMOSA,
):
    """Factory function of BipartiteSplitter instances.

    Utility function to simplificate BipartiteSplitter instantiation.
    Parameters may be set to a single value or a 2-valued
    tuple or list, to specify them for each axis.

    ax_min_samples_leaf represents [min_rows_leaf, min_cols_leaf]
    """
    if criteria is not None:
        bipartite_criterion = make_bipartite_criterion(
            criteria,
            bipartite_criterion_class,
            is_classification=is_classification,
            n_samples=n_samples,
            n_classes=n_classes,
            n_outputs=n_outputs,
        )
        criteria = [
            bipartite_criterion.criterion_rows,
            bipartite_criterion.criterion_cols,
        ]

    (
        max_features,
        ax_min_samples_leaf,
        ax_min_weight_leaf,
        splitters,
    ) = _duplicate_single_parameters(
        max_features,
        ax_min_samples_leaf,
        ax_min_weight_leaf,
        splitters,
    )

    random_state = check_random_state(random_state)

    for ax in range(2):
        # Make splitter
        if issubclass(splitters[ax], Splitter):
            if criteria is None:
                raise ValueError (
                    f"Since splitters[{ax}] is a class, criteria must "
                    "be provided."
                )
            if max_features[ax] is None:
                raise ValueError(
                    f"max_features[{ax}] must be provided if splitters"
                    f"[{ax}]={splitters[ax]} is a Splitter type.")
            splitters[ax] = splitters[ax](
                criterion=criteria[ax],
                max_features=max_features[ax],
                min_samples_leaf=ax_min_samples_leaf[ax],
                min_weight_leaf=ax_min_weight_leaf[ax],
                random_state=random_state,
            )
        elif criteria is not None:
            warnings.warn(
                f"Since splitters[{ax}] is not a class, the provided "
                f"{criteria[ax]=} is being ignored."
            )

    # Wrap splitters.
    return BipartiteSplitter(
        splitter_rows=splitters[0],
        splitter_cols=splitters[1],
        bipartite_criterion=bipartite_criterion,
        min_samples_leaf=min_samples_leaf,
        min_weight_leaf=min_weight_leaf,
    )


def make_bipartite_criterion(
    criteria,
    bipartite_criterion_class,
    *,
    is_classification=None,
    n_samples=None,
    n_classes=None,
    n_outputs=1,
):
    criteria, n_samples, n_outputs = _duplicate_single_parameters(
        criteria,
        n_samples,
        n_outputs,
    )

    if n_classes is None:
        n_classes = (None, None)
    else:
        if len(n_classes) != 2:
            raise ValueError(
                "A list with n_classes for each axis must be provided in the "
                "n_classes parameter."
            )

    for ax in range(2):
        criteria[ax] = check_criterion(
            criteria[ax],
            n_samples=n_samples[ax],
            n_outputs=n_outputs[ax],
            n_classes=n_classes[ax],
            is_classification=is_classification,
        )

    return bipartite_criterion_class(criteria[0], criteria[1])


# TODO: docs
#     """Unsupervised impurity is only used to decide between rows or columns.
# 
#     The split search takes into consideration only the labels, as usual, but
#     after the rows splitter and the columns splitter defines each one's split,
#     unsupervised information is used to decide between them, i.e. the final
#     impurity is semisupervised as in GMOSA, but the proxy improvement
#     only uses supervised data.
#     """
def make_bipartite_ss_splitter(
    splitters: list[Type[Splitter]],
    supervised_criteria=None,  # FIXME: validate, cannot pass None
    unsupervised_criteria=None,
    ss_criteria=None,
    supervision=0.5,
    update_supervision=None,  # TODO: for each axis
    n_features=None,
    n_samples=None,
    n_classes=None,
    is_classification=None,
    max_features=None,
    n_outputs=1,
    min_samples_leaf=1,
    min_weight_leaf=0.0,
    ax_min_samples_leaf=1,
    ax_min_weight_leaf=0.0,
    random_state=None,
    axis_decision_only=False,
    bipartite_criterion_class=GMOSA,
    ss_bipartite_criterion_class=BipartiteSemisupervisedCriterion,
):
    """Factory function of BipartiteSplitter instances with semisupervised criteria.

    Utility function to simplificate BipartiteSplitter instantiation.
    Parameters may be set to a single value or a 2-valued
    tuple or list, to specify them for each axis.
    """
    (
        max_features,
        ax_min_samples_leaf,
        ax_min_weight_leaf,
        splitters,
        supervised_criteria,
        unsupervised_criteria,
        ss_criteria,
    ) = _duplicate_single_parameters(
        max_features,
        ax_min_samples_leaf,
        ax_min_weight_leaf,
        splitters,
        supervised_criteria,
        unsupervised_criteria,
        ss_criteria,
    )

    random_state = check_random_state(random_state)

    bipartite_criterion = make_2dss_criterion(
        bipartite_criterion_class=bipartite_criterion_class,
        supervised_criteria=supervised_criteria,
        unsupervised_criteria=unsupervised_criteria,
        supervision=supervision,
        ss_criteria=ss_criteria,
        update_supervision=update_supervision,
        n_features=n_features,
        n_samples=n_samples,
        n_classes=n_classes,
        is_classification=is_classification,
        n_outputs=n_outputs,
        ss_bipartite_criterion_class=ss_bipartite_criterion_class,
    )

    if axis_decision_only:
        splitter_criteria = [
            bipartite_criterion.criterion_rows.supervised_criterion,
            bipartite_criterion.criterion_cols.supervised_criterion,
        ]
    else:
        splitter_criteria = [
            bipartite_criterion.criterion_rows,
            bipartite_criterion.criterion_cols,
        ]

    for ax in range(2):
        if not issubclass(splitters[ax], Splitter):
            raise TypeError

        splitters[ax] = splitters[ax](
            criterion=splitter_criteria[ax],
            max_features=max_features[ax],
            min_samples_leaf=ax_min_samples_leaf[ax],
            min_weight_leaf=ax_min_weight_leaf[ax],
            random_state=random_state,
        )

    return BipartiteSplitter(
        splitter_rows=splitters[0],
        splitter_cols=splitters[1],
        bipartite_criterion=bipartite_criterion,
        min_samples_leaf=min_samples_leaf,
        min_weight_leaf=min_weight_leaf,
    )


@validate_params(
    dict(
        supervision=[Interval(Real, 0.0, 1.0, closed="both")],
        supervised_criterion=[Criterion, type(Criterion)],
        unsupervised_criterion=[Criterion, type(Criterion)],
        is_classification=["boolean", None],
        n_outputs=[Interval(Integral, 1, None, closed="left"), None],
        n_features=[Interval(Integral, 1, None, closed="left"), None],
        n_samples=[Interval(Integral, 1, None, closed="left"), None],
        n_classes=[np.ndarray, None],
        update_supervision=[callable, None],
        ss_class=[type(SSCompositeCriterion), None],
    ),
    prefer_skip_nested_validation=True,
)
def make_semisupervised_criterion(
    *,
    supervision,
    supervised_criterion,
    unsupervised_criterion,
    is_classification=None,
    n_classes=None,
    n_outputs=None,
    n_features=None,
    n_samples=None,
    update_supervision=None,
    ss_class=None,
):
    supervised_criterion = check_criterion(
        criterion=supervised_criterion,
        n_outputs=n_outputs,
        n_samples=n_samples,
        n_classes=n_classes,
        is_classification=is_classification,
    )

    # Catch before check_criterion() to make it clear that it was the
    # n_features which was missing.
    if isinstance(unsupervised_criterion, type):
        if not n_features or not n_samples:
            raise ValueError(
                'If unsupervised_criterion is a class, one must provide both '
                f'n_features (received {n_features}) and n_samples ({n_samples}).'
            )

    unsupervised_criterion = check_criterion(
        criterion=unsupervised_criterion,
        n_outputs=n_features,
        n_samples=n_samples,
    )
    if isinstance(unsupervised_criterion, AxisCriterion):
        raise TypeError(
            f"{unsupervised_criterion=} cannot be an instance of AxisCriterion"
        )

    if ss_class is None:
        ss_class = SSCompositeCriterion

    return ss_class(
        supervision=supervision,
        supervised_criterion=supervised_criterion,
        unsupervised_criterion=unsupervised_criterion,
        update_supervision=update_supervision,
    )


def make_2dss_criterion(
    bipartite_criterion_class,
    supervised_criteria,
    unsupervised_criteria,
    supervision,
    ss_criteria=None,
    update_supervision=None,
    n_features=None,
    n_samples=None,
    n_classes=None,
    n_outputs=1,
    is_classification=None,
    ss_bipartite_criterion_class=BipartiteSemisupervisedCriterion,
):
    """Factory function of BipartiteSplitter instances with semisupervised criteria.

    Utility function to simplificate BipartiteSplitter instantiation.
    Parameters may be set to a single value or a 2-valued
    tuple or list, to specify them for each axis.
    """
    (
        n_samples,
        n_features,
        n_outputs,
        supervision,
        ss_criteria,
        supervised_criteria,
        unsupervised_criteria,
        update_supervision,
    ) = _duplicate_single_parameters(
        n_samples,
        n_features,
        n_outputs,
        supervision,
        ss_criteria,
        supervised_criteria,
        unsupervised_criteria,
        update_supervision,
    )

    if n_classes is None:
        n_classes = (None, None)

    # Make semi-supervised criteria
    for ax in range(2):
        if not isinstance(ss_criteria[ax], SSCompositeCriterion):
            ss_criteria[ax] = make_semisupervised_criterion(
                ss_class=ss_criteria[ax],
                supervision=supervision[ax],
                update_supervision=update_supervision[ax],
                supervised_criterion=supervised_criteria[ax],
                unsupervised_criterion=unsupervised_criteria[ax],
                n_outputs=n_outputs[ax],
                n_features=n_features[ax],
                n_samples=n_samples[ax],
                n_classes=n_classes[ax],
                is_classification=is_classification,
            )
        elif (
            supervised_criteria[ax] is not None
            or unsupervised_criteria[ax] is not None
        ):
            raise ValueError(
                "ss_criteria is not a type but supervised or unsupervised "
                f"criteria provided:\n  {supervised_criteria[ax]=}"
                f"\n  {unsupervised_criteria[ax]=}"
                f"\n  {ss_criteria[ax]=}"
            )
    
    return ss_bipartite_criterion_class(
        bipartite_criterion=bipartite_criterion_class(
            criterion_rows=ss_criteria[0],
            criterion_cols=ss_criteria[1],
        )
    )



# Author: Pedro Ilidio
# License: BSD 3 clause
from __future__ import annotations
from abc import ABCMeta, abstractmethod
import copy
from os import environ
from pathlib import Path
from typing import Callable
from urllib.parse import urlparse
from urllib.request import urlretrieve
import numpy as np
from sklearn.utils._param_validation import validate_params
from sklearn.datasets._base import _sha256

__all__ = [
    "get_data_home",
    "BaseFileLoader",
    "BaseRemoteFileLoader",
    "BipartiteDatasetLoader",
]

try:
    import pandas as pd
    Data = np.ndarray | pd.DataFrame | str
except ModuleNotFoundError:
    Data = np.ndarray | str


def get_data_home(data_home: str | Path | None = None) -> Path:
    """Return the path of the bipartite_learn data directory.  This folder is
    used by some large dataset loaders to avoid downloading the data several
    times.  By default the data directory is set to a folder named
    'bipartite_learn_data' in the user home folder. Alternatively, it can be
    set by the 'BIPARTITE_LEARN_DATA' environment variable or programmatically
    by giving an explicit folder path. The '~' symbol is expanded to the user
    home folder.  If the folder does not already exist, it is automatically
    created.
    Parameters
    ----------
    data_home : str, default=None
        The path to bipartite_learn data directory. If `None`, the default path
        is `~/bipartite_learn_data`.
    Returns
    -------
    data_home: Path
        The path to bipartite_learn data directory.
    """
    if data_home is None:
        data_home = environ.get(
            "BIPARTITE_LEARN_DATA", "~/bipartite_learn_data"
        )
    data_home = Path(data_home).expanduser()
    data_home.mkdir(parents=True, exist_ok=True)
    return data_home


class InvalidChecksumError(Exception):
    def __init__(self, path, actual=None, expected=None):
        self.path = path
        self.actual = actual
        self.expected = expected

    def __str__(self):
        msg = (
            f"{self.path}'s checksum differs from the expected value. File may"
            "be corrupted."
        )
        if self.expected is None or self.actual is None:
            return msg
        return msg + f"\n\tExpected: {self.expected}\n\tActual: {self.actual}"


class BaseFileLoader(metaclass=ABCMeta):
    """Abstract base class for file loader objects.
    Attributes
    ----------
    filepath : Path
        A Path object representing the location of the file to be loaded.
    description : str or BaseFileLoader
        A string describing the file being loaded or a BaseFileLoader instance 
        that provides additional details about the file.
    """

    def __init__(
        self,
        *,
        filepath: str | Path,
        description: str | BaseFileLoader = "",
    ):
        """Initialize the loader.
        Parameters
        ----------
        filepath : str or Path
            A Path object representing the location of the file to be loaded.
        description : str or BaseFileLoader, default=""
            A string describing the file being loaded or a BaseFileLoader
            instance that provides additional details about the file.
        """
        self.filepath: Path = Path(filepath)
        self.set_description(description)

    @property
    def local_path(self) -> Path:
        """The complete path to the local file."""
        return self.filepath.resolve()
    
    def set_description(self, description: str | BaseFileLoader) -> None:
        """Set the description of the file being loaded."""
        if not isinstance(description, (str, BaseFileLoader)):
            raise TypeError(
                f"{description=} must be either a string or a BaseFileLoader "
                "instance."
            )
        self.description = description

    def get_description(self) -> str:
        """Return the text description of the file being loaded."""
        if isinstance(self.description, BaseFileLoader):
            return self.description.load()
        return self.description
    
    @abstractmethod
    def load(self, as_frame: bool = False) -> Data:
        """Load the file and return its contents.
        Parameters
        ----------
        as_frame : bool, default=False
            Whether to return the data as a pandas DataFrame or not.
        Returns
        -------
        str or np.ndarray or pandas.DataFrame
            The contents of the file.
        """


class BaseRemoteFileLoader(BaseFileLoader, metaclass=ABCMeta):
    """Abstract base class for loaders of remote files.
    Attributes
    ----------
    url : str
        The URL address of the file to be fetched.
    filepath : Path
        A Path object representing the location of the file to be loaded,
        relative to ``self.base_dir``.
    description : str or BaseFileLoader
        A string describing the file being loaded or a BaseFileLoader instance
        that loads such string.
    base_dir : Path
        A Path object representing the root directory for ``self.filepath``.
    checksum : str or None
        The expected checksum of the file. If ``None``, integrity test is not
        performed after download.
    hash_function : Callable or None
        A function receiving the local filepath and returning its checksum.
        If ``None``, integrity test is not performed after download.
    """

    @validate_params(
        {
            "url": [str],
            "filepath": [str, Path, None],
            "description": [str, BaseFileLoader],
            "base_dir": [str, Path, None],
            "checksum": [str, None],
            "hash_function": [callable, None],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(
        self,
        *,
        url: str,
        filepath: str | Path | None = None,
        description: str | BaseRemoteFileLoader = "",
        base_dir: str | Path | None = None,
        checksum: str | None = None,
        hash_function: str | None = _sha256,
    ):
        """Initialize the loader.
        Parameters
        ----------
        url : str
            The URL address of the file to be fetched.
        filepath : str, Path or None, default=None
            A Path object representing the location of the file to be loaded,
            relative to ``self.base_dir``. If ``None``, the filename is inferred
            from the URL.
        description : str or BaseFileLoader, default=""
            A string describing the file being loaded or a BaseFileLoader instance
            that loads such string.
        base_dir : str, Path or None, default=None
            A Path object representing the root directory for ``self.filepath``.
            If ``None``, uses the current working directory.
        checksum : str or None, default=None
            The expected checksum of the file. If ``None``, integrity test is not
            performed after download.
        hash_function : Callable or None, default=sklearn.datasets._base._sha256
            A function receiving the local filepath and returning its checksum.
            If ``None``, integrity test is not performed after download.
        """
        self.url = url
        self.checksum = checksum
        self.hash_function = hash_function
        self.set_base_dir(base_dir)

        super().__init__(
            filepath=filepath or Path(urlparse(url).path).name,
            description=description,
        )

    @property
    def local_path(self) -> Path:
        """The complete path to the local file."""
        return (self.base_dir / self.filepath).resolve()
    
    def set_base_dir(self, base_dir: str | Path | None) -> None:
        """Set the root directory for ``self.filepath``."""
        self.base_dir = Path('.') if base_dir is None else Path(base_dir)

    def rebase_dir(self, new_base: str | Path) -> None:
        """Redirect the target download directory to be under new_base."""
        self.set_base_dir(Path(new_base) / self.base_dir)
   
    def set_description(self, description: str | BaseFileLoader) -> None:
        """Set the description of the file being loaded."""
        if isinstance(description, BaseRemoteFileLoader):
            # Copying avoid side effects when rebasing the directory
            description = copy.deepcopy(description)
            description.set_base_dir(self.base_dir)
        super().set_description(description)

    def download(self) -> Path:
        """Download the file to ``self.local_path`` and verify its integrity.
        """
        self.base_dir.mkdir(parents=True, exist_ok=True)
        urlretrieve(self.url, self.local_path)

        # If either checksum or hash_function was not provided, the integrity
        # test is not performed.
        if (self.checksum and self.hash_function) is not None:
            actual = self.hash_function(self.local_path)
            if actual != self.checksum:
                raise InvalidChecksumError(
                    self.local_path, actual=actual, expected=self.checksum
                )

        return self.local_path

    def clear_local(self) -> None:
        """Delete the local copy of the file."""
        self.local_path.unlink(missing_ok=True)
 
    def load(self, as_frame: bool = False) -> Data:
        """Load the local file if available, otherwise download it first.
        Parameters
        ----------
        as_frame : bool, default=False
            Whether to return the data as a pandas DataFrame or not.
        Returns
        -------
        str or np.ndarray or pandas.DataFrame
            The contents of the file.
        """
        if not self.local_path.exists():
            self.download()
        return self.load_local(as_frame)
    
    @abstractmethod
    def load_local(self, as_frame: bool = False) -> Data:
        """Load the local file and raise error if it is not available.
        Parameters
        ----------
        as_frame : bool, default=False
            Whether to return the data as a pandas DataFrame or not.
        Returns
        -------
        str or np.ndarray or pandas.DataFrame
            The contents of the file.
        """


class BipartiteDatasetLoader(BaseRemoteFileLoader):
    """Basic loader for bipartite datasets.

    This class groups data loaders for each file of a bipartite dataset.

    Attributes
    ----------
    X_loader : list[BaseRemoteFileLoader]
        List of remote file loaders for the feature matrices.
    y_loader : BaseRemoteFileLoader
        Remote file loader for the interaction matrix.
    filepath : Path
        Name for the local directory where the downloaded files will be stored.
        The directory will be created if it does not exist.
    base_dir : Path
        Base directory for the filepath. If ``None``, uses the current
        directory. 
    description : str or BaseRemoteFileLoader
        String description of the dataset or a file loader that loads it.
    """

    @validate_params(
        {
            "X_loader": [list],
            "y_loader": [BaseFileLoader],
            "filepath": [str, Path],
            "base_dir": [str, Path, None],
            "description": [str, BaseFileLoader],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(
        self,
        *,
        X_loader: list[BaseFileLoader],
        y_loader: BaseFileLoader,
        filepath: str | Path,
        base_dir: str | Path = None,
        description: str | BaseFileLoader = "",
    ):
        """Initializes a BipartiteDatasetLoader instance.
        Parameters
        ----------
        X_loader : list[BaseRemoteFileLoader]
            List of remote file loaders for the feature matrices.
        y_loader : BaseRemoteFileLoader
            Remote file loader for the interaction matrix.
        filepath : str or Path
            Name for a local directory where the downloaded files will be
            stored. The directory will be created if it does not exist.
        base_dir : str, Path or None, default=None
            Base directory for the filepath. If ``None``, uses the current
            directory. 
        description : str, BaseRemoteFileLoader, default=""
            String description of the dataset or a file loader that loads it.
        Raises
        ------
        NotADirectoryError
            If the local_path already exists and is not a directory.
        """
        # Deepcopy to avoid side effects when rebasing the directory
        self.y_loader = copy.deepcopy(y_loader)
        self.X_loader = [copy.deepcopy(X)  for X in X_loader]

        self.filepath = Path(filepath)
        self.set_base_dir(base_dir)

        if self.local_path.exists() and not self.local_path.is_dir():
            raise NotADirectoryError(f"{self.local_path=} is not a directory.")

        self.set_description(description)
    
    def set_base_dir(self, base_dir: str | Path | None) -> None:
        """Set the root directory for self.filepath and update loaders."""
        super().set_base_dir(base_dir)

        if isinstance(self.y_loader, BaseRemoteFileLoader):
            self.y_loader.set_base_dir(self.local_path)

        for X in self.X_loader:
            if isinstance(X, BaseRemoteFileLoader):
                X.set_base_dir(self.local_path)

    def download(self) -> Path:
        """Download the dataset files and verify their integrity."""
        if isinstance(self.y_loader, BaseRemoteFileLoader):
            self.y_loader.download()
        for X in self.X_loader:
            if isinstance(X, BaseRemoteFileLoader):
                X.download()
        return self.local_path
    
    def clear_local(self) -> None:
        """Delete the local copies of the files."""
        if isinstance(self.y_loader, BaseRemoteFileLoader):
            self.y_loader.clear_local()
        for X in self.X_loader:
            if isinstance(X, BaseRemoteFileLoader):
                X.clear_local()
        self.local_path.rmdir()
    
    def load_local(self, as_frame: bool = False) -> tuple[list[Data], Data]:
        """Load the local files and raise error if one is not available.
        Parameters
        ----------
        as_frame : bool, default=False
            Whether to return the data as a pandas DataFrames or numpy arrays.
        Returns
        -------
        X, y : tuple[list[Data], Data]
            Where Data is either pandas.DataFrame or numpy.ndarray.
            A tuple with the list of feature matrices for each axis and their
            corresponding interaction matrix.
        """
        if isinstance(self.y_loader, BaseRemoteFileLoader):
            y_data = self.y_loader.load_local(as_frame=as_frame) 
        else:
            y_data = self.y_loader.load(as_frame=as_frame)

        X_data = []
        for X in self.X_loader:
            if isinstance(X, BaseRemoteFileLoader):
                X_data.append(X.load_local(as_frame=as_frame))
            else:
                X_data.append(X.load(as_frame=as_frame))

        return X_data, y_data

    def load(self, as_frame: bool = False) -> tuple[list[Data], Data]:
        """Load the local files if available, otherwise download them first.
        Parameters
        ----------
        as_frame : bool, default=False
            Whether to return the data as a pandas DataFrames or numpy arrays.
        Returns
        -------
        X, y : tuple[list[Data], Data]
            Where Data is either pandas.DataFrame or numpy.ndarray.
            A tuple with the list of feature matrices for each axis and their
            corresponding interaction matrix.
        """
        return super().load(as_frame=as_frame)


#! /usr/bin/env python
# Copyright (C) 2022 Pedro Ilídio <ilidio@alumni.usp.br>
# License: 3-clause BSD

from setuptools import setup, Extension, find_packages
from pathlib import Path
from Cython.Build import cythonize
import numpy

PATH_ROOT = Path(__file__).parent
README = (PATH_ROOT/"README.md").read_text()
# Get __version__ from _version.py
exec((PATH_ROOT/"bipartite_learn/_version.py").read_text())
VERSION = __version__

extensions = [Extension("bipartite_learn.tree.*", ["bipartite_learn/tree/*.pyx"])]

setup(
    name='bipartite_learn',
    version=VERSION,
    description='Machine learning estimators for bipartite data.',
    include_dirs=[numpy.get_include()],
    long_description=README,
    long_description_content_type="text/markdown",
    url='http://github.com/pedroilidio/bipartite_learn',
    author='Pedro Ilídio',
    author_email='pedrilidio@gmail.com',
    license='new BSD',
    packages=find_packages(),
    zip_safe=False,

    install_requires=[
        'cython==0.29.33',
        'scikit-learn==1.3.0',
        'numpy>=1.22.2',
        'imbalanced-learn==0.9.1',
    ],
    extras_require={
        'docs': ['sphinx', 'pydata-sphinx-theme'],
    },

    ext_modules=cythonize(
        extensions,
        language_level="3",
    ),
)



# distutils: language = c++
# cython: boundscheck=False
from cpython cimport Py_INCREF, PyObject, PyTypeObject
from libc.stdint cimport INTPTR_MAX
import struct
import warnings
import numpy as np
cimport numpy as cnp
cnp.import_array()

from scipy.sparse import issparse

cdef extern from "numpy/arrayobject.h":
    object PyArray_NewFromDescr(PyTypeObject* subtype, cnp.dtype descr,
                                int nd, cnp.npy_intp* dims,
                                cnp.npy_intp* strides,
                                void* data, int flags, object obj)
    int PyArray_SetBaseObject(cnp.ndarray arr, PyObject* obj)

cdef extern from "<stack>" namespace "std" nogil:
    cdef cppclass stack[T]:
        ctypedef T value_type
        stack() except +
        bint empty()
        void pop()
        void push(T&) except +  # Raise c++ exception for bad_alloc -> MemoryError
        T& top()

# =============================================================================
# Types and constants
# =============================================================================

from numpy import float32 as DTYPE
from numpy import float64 as DOUBLE

cdef double INFINITY = np.inf
cdef double EPSILON = np.finfo('double').eps

# Some handy constants (BestFirstTreeBuilder)
cdef int IS_FIRST = 1
cdef int IS_NOT_FIRST = 0
cdef int IS_LEFT = 1
cdef int IS_NOT_LEFT = 0

TREE_LEAF = -1
TREE_UNDEFINED = -2
cdef SIZE_t _TREE_LEAF = TREE_LEAF
cdef SIZE_t _TREE_UNDEFINED = TREE_UNDEFINED
cdef SIZE_t INITIAL_STACK_SIZE = 10


cdef class BipartiteTreeBuilder:
    """Interface for different tree building strategies."""

    cpdef build(
        self,
        Tree tree,
        object X,
        const DOUBLE_t[:, :] y,
        const DOUBLE_t[:] sample_weight=None,
    ):
        """Build a decision tree from the training set (X, y)."""

    cdef inline _check_input(
        self,
        object X,
        const DOUBLE_t[:, :] y,
        const DOUBLE_t[:] sample_weight,
    ):
        """Check input dtype, layout and format.

        Applies the same processing as sklearn's TreeBuilder for each X[ax]
        feature matrix in X.
        """
        for Xi in X:
            if issparse(Xi):
                raise NotImplementedError("No support for sparse matrices.")

            elif (
                Xi.dtype != DTYPE
                # or not Xi.flags.contiguous
            ):
                # FIXME: fortran contiguous
                raise TypeError("X matrices must be float32.")

        if (
            sample_weight is not None
            and (
                sample_weight.base.dtype != DOUBLE
                or not sample_weight.base.flags.contiguous
            )
        ):
            sample_weight = np.asarray(sample_weight, dtype=DOUBLE, order="C")
            warnings.warn(
                "A contiguous version of sample_weight was created. Unecessary"
                " memory load is expected if the tree composes an ensemble."
                " Passing an already contiguous sample_weight is recommended."
            )

        return X, y, sample_weight


# Depth first builder ---------------------------------------------------------
# A record on the stack for depth-first tree growing
cdef struct StackRecord2D:
    SIZE_t start_row
    SIZE_t start_col
    SIZE_t end_row
    SIZE_t end_col
    SIZE_t depth
    SIZE_t parent
    bint is_left
    double impurity
    SIZE_t n_constant_row_features
    SIZE_t n_constant_col_features


cdef class BipartiteDepthFirstTreeBuilder(BipartiteTreeBuilder):
    """Build a decision tree in depth-first fashion, from 2D training data.

    It adds minor changes to sklearn's DepthfirstTreeBuilder, essentially
    in storing two-values start/end node positions and managing getting the
    new ones after each split.

    `X` is now a list with 2 matrices, for the feature table of row and column
    instances respectively. `y` is a 2 dimensional ndarray representing labels
    of interaction of each row with each column. Mulit-output is not yet imp-
    lemented.
    """
    # TODO: define separate methods for split -> node data conversion and
    # evaluating stopping criteria.
    def __cinit__(
        self,
        BipartiteSplitter splitter,
        SIZE_t min_samples_split,
        SIZE_t min_samples_leaf,
        double min_weight_leaf,
        SIZE_t max_depth,
        double min_impurity_decrease,
        # Bipartite parameters
        SIZE_t min_rows_split=1,  # Not 2, to still allow split on the other axis
        SIZE_t min_cols_split=1,
        SIZE_t min_cols_leaf=1,
        SIZE_t min_rows_leaf=1,
        double min_row_weight_leaf=0.0,
        double min_col_weight_leaf=0.0,
    ):
        self.splitter = splitter
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_leaf = min_weight_leaf
        self.max_depth = max_depth
        self.min_impurity_decrease = min_impurity_decrease
        # Bipartite parameters
        self.min_rows_split = min_rows_split
        self.min_cols_split = min_cols_split
        self.min_rows_leaf = min_rows_leaf
        self.min_cols_leaf = min_cols_leaf
        self.min_row_weight_leaf = min_weight_leaf
        self.min_col_weight_leaf = min_weight_leaf

    cpdef build(
        self,
        Tree tree,
        object X,
        const DOUBLE_t[:, :] y,
        const DOUBLE_t[:] sample_weight=None,
    ):
        """Build a decision tree from the training set (X, y)."""
        # check input
        X, y, sample_weight = self._check_input(X, y, sample_weight)

        # Initial capacity
        cdef int init_capacity

        if tree.max_depth <= 10:
            init_capacity = <int> (2 ** (tree.max_depth + 1)) - 1
        else:
            init_capacity = 2047

        tree._resize(init_capacity)

        # Parameters
        cdef BipartiteSplitter splitter = self.splitter
        cdef SIZE_t max_depth = self.max_depth
        cdef SIZE_t min_samples_leaf = self.min_samples_leaf
        cdef SIZE_t min_samples_split = self.min_samples_split
        cdef double min_weight_leaf = self.min_weight_leaf
        cdef double min_impurity_decrease = self.min_impurity_decrease

        # Bipartite parameters
        cdef SIZE_t min_rows_leaf = self.min_rows_leaf
        cdef SIZE_t min_cols_leaf = self.min_cols_leaf
        cdef SIZE_t min_rows_split = self.min_rows_split
        cdef SIZE_t min_cols_split = self.min_cols_split
        cdef double min_row_weight_leaf = self.min_row_weight_leaf
        cdef double min_col_weight_leaf = self.min_col_weight_leaf

        splitter.init(X, y, sample_weight)

        cdef SIZE_t[2] start, start_left, start_right
        cdef SIZE_t[2] end, end_left, end_right
        cdef SIZE_t depth
        cdef SIZE_t parent
        cdef bint is_left
        cdef SIZE_t n_node_samples = splitter.n_samples
        cdef SIZE_t n_node_rows = splitter.n_rows
        cdef SIZE_t n_node_cols = splitter.n_cols
        cdef double weighted_n_node_samples[3]  # total samples, rows, columns
        cdef MultipartiteSplitRecord split
        cdef SIZE_t node_id

        cdef double impurity = INFINITY
        cdef SIZE_t[2] n_constant_features
        cdef bint is_leaf
        cdef bint first = 1
        cdef SIZE_t max_depth_seen = -1
        cdef int rc = 0

        cdef stack[StackRecord2D] builder_stack
        cdef StackRecord2D stack_record

        # Recursive partition (without actual recursion)
        with nogil:
            # push root node onto stack
            builder_stack.push({
                "start_row": 0,
                "start_col": 0,
                "end_row": splitter.splitter_rows.n_samples,
                "end_col": splitter.splitter_cols.n_samples,
                "depth": 0,
                "parent": _TREE_UNDEFINED,
                "is_left": 0,
                "impurity": INFINITY,
                "n_constant_row_features": 0,
                "n_constant_col_features": 0,
            })

            while not builder_stack.empty():
                stack_record = builder_stack.top()
                builder_stack.pop()

                start[0] = stack_record.start_row
                start[1] = stack_record.start_col
                end[0] = stack_record.end_row
                end[1] = stack_record.end_col
                depth = stack_record.depth
                parent = stack_record.parent
                is_left = stack_record.is_left
                impurity = stack_record.impurity
                n_constant_features[0] = stack_record.n_constant_row_features
                n_constant_features[1] = stack_record.n_constant_col_features
                n_node_rows = end[0] - start[0]
                n_node_cols = end[1] - start[1]
                n_node_samples = n_node_rows * n_node_cols

                splitter.node_reset(start, end, weighted_n_node_samples)

                is_leaf = (
                    depth >= max_depth
                    or n_node_samples < min_samples_split
                    or n_node_samples < 2 * min_samples_leaf
                    or weighted_n_node_samples[0] < 2 * min_weight_leaf
                    # Bipartite parameters
                    or n_node_rows < min_rows_split
                    or n_node_cols < min_cols_split
                    # Not times 2, to still allow splitting in the other axis
                    or n_node_rows < min_rows_leaf 
                    or n_node_cols < min_cols_split
                    or weighted_n_node_samples[1] < 2 * min_row_weight_leaf
                    or weighted_n_node_samples[2] < 2 * min_col_weight_leaf
                )

                if first:
                    impurity = splitter.node_impurity()
                    first = 0

                # impurity == 0 with tolerance due to rounding errors
                is_leaf = is_leaf or impurity <= EPSILON  # TODO

                if not is_leaf:
                    splitter.node_split(impurity, &split, n_constant_features)
                    # If EPSILON=0 in the below comparison, float precision
                    # issues stop splitting, producing trees that are
                    # dissimilar to v0.18
                    is_leaf = (
                        split.split_record.pos >= end[split.axis]
                        or (
                            split.split_record.improvement + EPSILON
                            < min_impurity_decrease
                        )
                    )

                node_id = tree._add_node(
                    parent,
                    is_left,
                    is_leaf,
                    split.split_record.feature,
                    split.split_record.threshold,
                    impurity,
                    n_node_samples,
                    weighted_n_node_samples[0],
                    False,  # TODO: missing values support
                )

                if node_id == INTPTR_MAX:
                    rc = -1
                    break

                # Store value for all nodes, to facilitate tree/model
                # inspection and interpretation
                splitter.node_value(tree.value + node_id * tree.value_stride)

                if not is_leaf:
                    # FIXME: review.
                    ## https://github.com/cython/cython/pull/1663
                    # start_left[:] = start[:]
                    # end_left[:] = end[:]

                    # for ax in range(n_dimensions):
                    #     start_left[ax] = start[ax]
                    #     start_right[ax] = start[ax]
                    #     end_left[ax], end_right[ax] = end[ax], end[ax]

                    start_left[0], start_right[0] = start[0], start[0]
                    start_left[1], start_right[1] = start[1], start[1]
                    end_left[0], end_right[0] = end[0], end[0]
                    end_left[1], end_right[1] = end[1], end[1]

                    # Setting new nodes coordinates.
                    # These lines are why we had to rewrite this whole method
                    # to develop the 2-dimensional version.
                    start_right[split.axis] = split.split_record.pos
                    end_left[split.axis] = split.split_record.pos

                    # Push right child on stack
                    builder_stack.push({
                        "start_row": start_right[0],
                        "start_col": start_right[1],
                        "end_row": end_right[0],
                        "end_col": end_right[1],
                        "depth": depth + 1,
                        "parent": node_id,
                        "is_left": 0,
                        "impurity": split.split_record.impurity_right,
                        "n_constant_row_features": n_constant_features[0],
                        "n_constant_col_features": n_constant_features[1],
                    })

                    # Push left child on stack
                    builder_stack.push({
                        "start_row": start_left[0],
                        "start_col": start_left[1],
                        "end_row": end_left[0],
                        "end_col": end_left[1],
                        "depth": depth + 1,
                        "parent": node_id,
                        "is_left": 1,
                        "impurity": split.split_record.impurity_left,
                        "n_constant_row_features": n_constant_features[0],
                        "n_constant_col_features": n_constant_features[1],
                    })

                if depth > max_depth_seen:
                    max_depth_seen = depth

            if rc >= 0:
                rc = tree._resize_c(tree.node_count)

            if rc >= 0:
                tree.max_depth = max_depth_seen
        if rc == -1:
            raise MemoryError()


cdef class BipartiteBestFirstTreeBuilder(BipartiteTreeBuilder):
    ...  # TODO



import joblib
import pytest
from pathlib import Path
import numpy as np
from sklearn.tree import DecisionTreeRegressor, export_text
from sklearn.ensemble import RandomForestRegressor
from sklearn.base import clone
from sklearn.model_selection._validation import _score
from sklearn.metrics import check_scoring
from bipartite_learn.wrappers import GlobalSingleOutputWrapper
from bipartite_learn.melter import melt_multipartite_dataset
from .utils.make_examples import make_interaction_regression
from .utils.test_utils import stopwatch

with stopwatch('Making dataset'):
    XX, Y, X, y = make_interaction_regression(
        return_molten=True,
        n_samples=(50, 60),
        n_features=(10, 9),
        random_state=0,
        max_target=1.0,
    )
    # Make it binary  # TODO: specialized data generation function
    y_mean = Y.mean()
    Y = (Y > y_mean).astype(int)
    y = (y > y_mean).astype(int)
    # Ensure negative majority
    if Y.mean() > 0.5:
        Y, y = 1-Y, 1-y


@pytest.fixture(params=range(10))
def random_state(request):
    return request.param


@pytest.fixture(params=[1, 0.1])
def min_samples_leaf(request):
    return request.param


def test_gso_wrapper_bipartite_subsampling(random_state, min_samples_leaf):
    Y_ = (Y == 1).astype(int)
    y_ = (y == 1).astype(int)

    tree = DecisionTreeRegressor(
        min_samples_leaf=min_samples_leaf,
        random_state=random_state,
    )
    tree_bipartite = GlobalSingleOutputWrapper(clone(tree))

    tree_bipartite.fit(XX, Y_)
    tree.fit(X, y_)

    assert export_text(tree_bipartite.estimator_) == export_text(tree)


def _test_pickling(obj, tmp_path):
    joblib.dump(obj, tmp_path/'test.pickle')
    obj_loaded = joblib.load(tmp_path/'test.pickle')
    obj.get_params()
    obj_loaded.get_params()


def test_wrapped_tree_pickling(tmp_path):
    tree = DecisionTreeRegressor(
        min_samples_leaf=30,
        random_state=0,
    )
    tree_bipartite = GlobalSingleOutputWrapper(clone(tree))

    _test_pickling(tree, tmp_path)
    _test_pickling(tree_bipartite, tmp_path)


def test_wrapped_forest_pickling(tmp_path):
    forest = RandomForestRegressor(
        min_samples_leaf=30,
        random_state=0,
    )
    forest_bipartite = GlobalSingleOutputWrapper(
        clone(forest)
    )

    _test_pickling(forest, tmp_path)
    _test_pickling(forest_bipartite, tmp_path)


def test_score():
    forest = RandomForestRegressor(
        min_samples_leaf=30,
        random_state=0,
    )
    forest_bipartite = GlobalSingleOutputWrapper(
        clone(forest),
    )
    forest_bipartite.fit(XX, Y)

    score = _score(
        forest_bipartite, X, y,
        check_scoring(forest_bipartite, "roc_auc"),
        error_score="raise",
    )
    print("Score:", score)


def test_gso():
    from sklearn.ensemble import RandomForestClassifier
    from imblearn.under_sampling import RandomUnderSampler
    from bipartite_learn.datasets import NuclearReceptorsLoader
    X, y = NuclearReceptorsLoader().load()  # X is a list of two matrices

    bipartite_clf = GlobalSingleOutputWrapper(
        estimator=RandomForestClassifier(),
        under_sampler=RandomUnderSampler(),
    )
    bipartite_clf.fit(X, y)
    assert bipartite_clf.score(X, y) > 0.5


def test_gmo():
    from bipartite_learn.datasets import NuclearReceptorsLoader
    from bipartite_learn.wrappers import LocalMultiOutputWrapper
    from sklearn.svm import SVC
    from sklearn.neighbors import KNeighborsClassifier
    from sklearn.multioutput import MultiOutputClassifier
    
    X, y = NuclearReceptorsLoader().load()  # X is a list of two matrices
    bipartite_clf = LocalMultiOutputWrapper(
        primary_rows_estimator=MultiOutputClassifier(SVC()),
        primary_cols_estimator=MultiOutputClassifier(SVC()),
        secondary_rows_estimator=KNeighborsClassifier(),
        secondary_cols_estimator=KNeighborsClassifier(),
        combine_predictions_func=np.max,
    )
    bipartite_clf.fit(X, y)
    assert bipartite_clf.score(X, y) > 0.5


import numpy as np
from sklearn.utils.validation import check_random_state
from sklearn.utils._param_validation import validate_params
from bipartite_learn.base import (
    BaseBipartiteEstimator,
    BaseMultipartiteSampler,
)
from bipartite_learn.utils import _X_is_multipartite

__all__ = [
    "row_cartesian_product",
    "melt_multipartite_dataset",
    "BipartiteMelter",
    "MultiPartiteMelter",
]


def row_cartesian_product(X):
    """Row cartesian product of 2D arrays in X.

    Pick one row from each of the 2D arrays in X, in their presented order, and
    concatenate them. Repeat. Return a 2D array where its rows are all the
    possible combinations of rows in X.

    Parameters
    ----------
    X : list-like of 2D np.ndarrays

    Returns
    -------
    result : 2D np.ndarray
        Cartesian product of X's 2d arrays, row-wise.
    """
    lengths = [Xi.shape[0] for Xi in X]

    # Xi will repeat once for every combination of samples ahead (Xj if j>i)
    # and tile once for every combination of samples before (Xj if j<i).
    n_repeat_tile = [
        (np.prod(lengths[(i+1):]), np.prod(lengths[:i]))
        for i in range(len(X))
    ]

    return np.hstack([
        np.tile(Xi.repeat(n_repeat, axis=0), (n_tile, 1))
        for Xi, (n_repeat, n_tile) in zip(X, n_repeat_tile)
    ])


@validate_params(
    {'X': [list, tuple], 'y': ['array-like', None]},
    prefer_skip_nested_validation=True,
)
def melt_multipartite_dataset(X, y=None):
    """Melt bipartite input.
    
    If X is a list of Xi feature matrices, one for each bipartite group,
    convert it to traditional data format by generating concatenations of
    rows from X[0] with rows from X[1].
    """
    if not _X_is_multipartite(X):
        raise ValueError("Tried to melt monopartite dataset.")

    X = row_cartesian_product(X) 

    if y is None:
        return X, y

    return X, y.reshape(-1, 1)


class BipartiteMelter(BaseMultipartiteSampler, BaseBipartiteEstimator):
    """Convert a bipartite dataset to a simpler global-single output format.

    Convert a bipartite interaction problem, where there are two feature
    matrices in X (one for each axis) and an interaction matrix y to a simpler
    usual format where each sample is a combination of samples from X[0] and
    X[1].

    Slightly faster than MultipartiteMelter.
    """

    def _fit_resample(self, X, y):
        X1, X2 = X
        melted_X = np.hstack([
            np.repeat(X1, X2.shape[0], axis=0),
            np.tile(X2, (X1.shape[0], 1))
        ])
        return melted_X, y.reshape(-1, 1)


class MultipartiteMelter(BaseMultipartiteSampler):
    """Convert a multipartite dataset to a simpler global-single output format.
    """

    def _fit_resample(self, X, y):
        return melt_multipartite_dataset(X, y)


"""
The :mod:`bipartite_learn.utils` module includes various utilities.
"""
# TODO: reallocate as much as we can.

from __future__ import annotations
import inspect
import pkgutil
import numpy as np
from importlib import import_module
from operator import itemgetter
from pathlib import Path
from typing import Any, Sequence
from sklearn.utils import check_array
from sklearn.utils.validation import check_symmetric
from sklearn.utils.validation import _check_sample_weight
from .discovery import all_estimators

__all__ = [
    "all_estimators",
    "check_multipartite_params",
    "check_simmilarity_matrix",
]


def _check_multipartite_sample_weight(sample_weight, X, **kwargs):
    if not isinstance(sample_weight, np.ndarray):  # Number or None
        row_weight = col_weight = sample_weight
    else:
        n_rows = X[0].shape[0]
        row_weight = sample_weight[:n_rows]
        col_weight = sample_weight[n_rows:]

    row_weight = _check_sample_weight(row_weight, X[0], **kwargs)
    col_weight = _check_sample_weight(col_weight, X[1], **kwargs)

    return np.hstack([row_weight, col_weight])


def _X_is_multipartite(X):
    # TODO: find a better way of deciding.
    return isinstance(X, (tuple, list))


def check_similarity_matrix(
    X,
    *,
    check_symmetry=True,
    symmetry_tol=1e-10,
    symmetry_warning=True,
    symmetry_exception=False,
    **check_array_args,
):
    X = check_array(X, **check_array_args)
    eps = np.finfo(X.dtype).eps

    if (X > 1. + eps).any() or (X < -eps).any():
        raise ValueError(
            "Similarity values must be between 0 and 1 "
            f"(inclusive). Found {X.min()=} and {X.max()=}.",
        )

    if check_symmetry:
        return check_symmetric(
            X,
            tol=symmetry_tol,
            raise_warning=symmetry_warning,
            raise_exception=symmetry_exception,
        )
    return X


def check_multipartite_params(*params, k=2):
    new_params = []

    for p in params:
        if isinstance(p, Sequence) and not isinstance(p, str):
            if len(p) != k:
                raise ValueError(
                    f"Parameter {p} was required to have length={k}")
            new_params.append(p)
        else:
            new_params.append([p for _ in range(k)])
    
    return new_params[0] if len(params) == 1 else new_params


def _partiteness_name(k: int) -> str:
    if k < 1:
        raise ValueError("Invalid partiteness.")
    elif k == 1:
        return "monopartite"
    elif k == 2:
        return "bipartite"
    elif k == 3:
        return "tripartite"
    else:
        return f"{k}-partite"


def check_partiteness(
    X=None,
    y=None,
    *,
    partiteness: None | int = None,
    estimator: None | str | Any = None,
):
    if estimator is not None and not isinstance(estimator, str):
        if hasattr(estimator, '_partiteness'):
            partiteness = estimator._partiteness
            if partiteness is None:  # Estimator accepts any partiteness
                return
        elif partiteness is None:
            raise ValueError("If no partiteness provided, estimator must have "
                             "a 'partiteness' attribute.")
    elif partiteness is None:
        raise ValueError(
            "Either partiteness or an estimator instance must be provided."
        )

    kpartite = _partiteness_name(partiteness)

    if estimator is None:
        but_text = f"but {kpartite} input was expected."
    elif isinstance(estimator, str):
        but_text = f"but {estimator} expects {kpartite} input."
    else:
        but_text = f"but {type(estimator).__name__} expects {kpartite} input."

    if X is None and y is None:
        raise ValueError("Either X or y must be provided.")

    if X is not None:
        len_X = len(X)
        if len_X != partiteness:
            raise ValueError(
                f"X is {_partiteness_name(len_X)} ({len(X)=}) " + but_text
            )
    if y is not None:
        if y.ndim != partiteness:
            raise ValueError(
                f"y is {_partiteness_name(y.ndim)} ({y.ndim=}) " + but_text
            )


# TODO: docs
# TODO: KNN sampler to be used by NRLMF
# TODO: Beta correction kernel aggregator
# TODO: MultipartiteSamplerWrapper
from __future__ import annotations
from numbers import Number
import numpy as np
from sklearn.gaussian_process.kernels import RBF
from ..base import BaseMultipartiteSampler


class DTHybridSampler(BaseMultipartiteSampler):
    def __init__(self, lamb=.5, alpha=.5):
        self.lamb = lamb
        self.alpha = alpha

    def _fit_resample(self, X, y):
        y = y.astype(float)  # TODO: Use sklearn utilities.
        S_row_net = y @ y.T  # Rows network similarity
        k_row = np.diag(S_row_net).copy()  # Row nodes degree
        # If the row makes no interactions, set the similarity to 0
        # (bellow we will divide the similarity by k_row)
        # TODO: raise warning.
        k_row[k_row == 0] = np.inf

        # k_row_total[i, j] = k_row[i] + k_row[j]
        k_row_total = k_row[:, np.newaxis] * k_row

        # k_row_weighted[i, j] = lambda**k_row[i] + k_row[j]**(1-lambda)
        k_row_weighted = (
            k_row[:, np.newaxis]**self.lamb * k_row**(1-self.lamb)
        )

        k_col = y.sum(axis=0)  # Column nodes degree
        # If the column makes no interactions, set the similarity to 0
        k_col[k_col == 0] = np.inf

        # Row similarity based on column chemical similarity
        S_row_from_column = y @ X[1] @ y.T / k_row_total

        S = self.alpha * X[0] + (1-self.alpha) * S_row_from_column
        Gamma = S / k_row_weighted

        W = Gamma * ((y / k_col) @ y.T)

        return X, W @ y
    


# cython: boundscheck=False
import numpy as np
from libc.stdlib cimport malloc, calloc, free, realloc
from libc.string cimport memset
from sklearn.tree._criterion cimport RegressionCriterion, Criterion
from ._axis_criterion cimport AxisCriterion, AxisClassificationCriterion


class InvalidAxisError(ValueError):
    def __init__(self, axis: int | None = None):
        self.axis = axis
    def __str__(self):
        return (
            "'axis' parameter can only be 0 or 1"
            f", not {self.axis}." if self.axis is not None else "."
        )


cdef class BipartiteCriterion:
    """Abstract base class."""

    cdef int init(
        self,
        const DTYPE_t[:, :] X_rows,
        const DTYPE_t[:, :] X_cols,
        const DOUBLE_t[:, :] y,
        DOUBLE_t[:] row_weights,
        DOUBLE_t[:] col_weights,
        double weighted_n_rows,
        double weighted_n_cols,
        const SIZE_t[:] row_indices,
        const SIZE_t[:] col_indices,
        SIZE_t[2] start,
        SIZE_t[2] end,
    ) nogil except -1:
        return -1

    cdef void node_value(self, double* dest) nogil:
        pass

    cdef double node_impurity(self) nogil:
        pass

    cdef void children_impurity(
            self,
            double* impurity_left,
            double* impurity_right,
            SIZE_t axis,
    ) nogil:
        pass

    cdef double impurity_improvement(
        self,
        double impurity_parent,
        double impurity_left,
        double impurity_right,
        SIZE_t axis,
    ) nogil:
        """The final value to express the split quality. 
        """
    

cdef class GMO(BipartiteCriterion):
    """Applies Predictive Bi-Clustering Trees method.

    See [Pliakos _et al._, 2018](https://doi.org/10.1007/s10994-018-5700-x).
    """

    def __reduce__(self):
        return (
            type(self),
            (self.criterion_rows, self.criterion_cols),
            self.__getstate__(),
        )

    def __getstate__(self):
        return {}

    def __cinit__(self):
        # Default values
        self.sq_sum_total = 0.0

        self.start[0] = 0
        self.start[1] = 0
        self.end[0] = 0
        self.end[1] = 0

        self.weighted_n_node_samples = 0.0
        self.weighted_n_node_rows = 0.0
        self.weighted_n_node_cols = 0.0

        self.sum_total = np.zeros(self.n_outputs, dtype=np.float64)

    def __init__(
        self,
        AxisCriterion criterion_rows,
        AxisCriterion criterion_cols,
    ):
        # Objects must be set here, to ensure they are fully initialised
        self.criterion_rows = criterion_rows
        self.criterion_cols = criterion_cols

        self.n_rows = self.criterion_rows.n_samples
        self.n_cols = self.criterion_cols.n_samples
        self.n_outputs_rows = self.criterion_rows.n_outputs
        self.n_outputs_cols = self.criterion_cols.n_outputs

        # FIXME: should not validate arguments here.
        if isinstance(criterion_rows, AxisClassificationCriterion):
            if not isinstance(criterion_cols, AxisClassificationCriterion):
                raise TypeError(
                    "None or both axes criteria must be "
                    "AxisClassificationCriterion."
                )
            self.max_n_classes = (
                (<AxisClassificationCriterion>criterion_cols).max_n_classes
            )
        else:
            self.max_n_classes = 1

        self.n_outputs = self.n_outputs_rows + self.n_outputs_cols

    cdef int init(
        self,
        const DTYPE_t[:, :] X_rows,
        const DTYPE_t[:, :] X_cols,
        const DOUBLE_t[:, :] y,
        DOUBLE_t[:] row_weights,
        DOUBLE_t[:] col_weights,
        double weighted_n_rows,
        double weighted_n_cols,
        const SIZE_t[:] row_indices,
        const SIZE_t[:] col_indices,
        SIZE_t[2] start,
        SIZE_t[2] end,
    ) nogil except -1:
        """This function adapts RegressionCriterion.init to 2D data."""

        # Initialize fields
        self.X_rows = X_rows
        self.X_cols = X_cols
        self.y = y
        self.row_weights = row_weights
        self.col_weights = col_weights
        self.weighted_n_rows = weighted_n_rows
        self.weighted_n_cols = weighted_n_cols
        self.weighted_n_samples = weighted_n_rows * weighted_n_cols
        self.row_indices = row_indices
        self.col_indices = col_indices

        self.start[0], self.start[1] = start[0], start[1]
        self.end[0], self.end[1] = end[0], end[1]

        cdef SIZE_t n_node_rows = end[0] - start[0]
        cdef SIZE_t n_node_cols = end[1] - start[1]

        self.criterion_rows.axis_init(
            y=self.y,
            sample_weight=self.row_weights,
            col_weights=self.col_weights,
            weighted_n_samples=self.weighted_n_rows,
            weighted_n_cols=self.weighted_n_cols,
            sample_indices=self.row_indices,
            col_indices=self.col_indices,
            start=self.start[0],
            end=self.end[0],
            start_col=self.start[1],
            end_col=self.end[1],
        )
        self.criterion_cols.axis_init(
            y=self.y.T,
            sample_weight=self.col_weights,
            col_weights=self.row_weights,
            weighted_n_samples=self.weighted_n_cols,
            weighted_n_cols=self.weighted_n_rows,
            sample_indices=self.col_indices,
            col_indices=self.row_indices,
            start=self.start[1],
            end=self.end[1],
            start_col=self.start[0],
            end_col=self.end[0],
        )

        # Will be used by TreeBuilder as stopping criteria.
        self.weighted_n_node_rows = self.criterion_rows.weighted_n_node_samples
        self.weighted_n_node_cols = self.criterion_cols.weighted_n_node_samples

        # Will further be used by the BipartiteSplitter to set the Tree object
        self.weighted_n_node_samples = (
            self.weighted_n_node_rows * self.weighted_n_node_cols
        )

        return 0

    cdef void node_value(self, double* dest) nogil:
        """Copy the value (prototype) of node sample_indices into dest.
        """
        self.criterion_cols.node_value(dest)
        self.criterion_rows.node_value(
            dest + self.n_outputs_cols * self.max_n_classes
        )

    cdef double node_impurity(self) nogil:
        """Return the impurity of the current node.

        In scikit-learn trees it is only used at the root node.
        """
        # Should be equal among axes.
        return self.criterion_rows.node_impurity()

    cdef void children_impurity(
            self,
            double* impurity_left,
            double* impurity_right,
            SIZE_t axis,
    ) nogil:
        (<AxisCriterion> self._get_criterion(axis)).children_impurity(
            impurity_left,
            impurity_right,
        )

    cdef double impurity_improvement(
        self,
        double impurity_parent,
        double impurity_left,
        double impurity_right,
        SIZE_t axis,
    ) nogil:
        """The final value to express the split quality. 
        """
        return (<AxisCriterion>self._get_criterion(axis)).impurity_improvement(
            impurity_parent,
            impurity_left,
            impurity_right,
        )
 
    cdef void* _get_criterion(self, SIZE_t axis) nogil:
        if axis == 0:
            return <void*> self.criterion_rows
        elif axis == 1:
            return <void*> self.criterion_cols
        else:
            with gil:
                raise InvalidAxisError(axis)


cdef class GMOSA(GMO):
    def __init__(
        self,
        AxisCriterion criterion_rows,
        AxisCriterion criterion_cols,
    ):
        super().__init__(criterion_rows, criterion_cols)
        self.n_outputs = 1

    cdef void node_value(self, double* dest) nogil:
        """Copy the value (prototype) of node sample_indices into dest.
        """
        self.criterion_rows.total_node_value(dest)